/*  Screem:  screem-tagtree.c
 *
 *  the tag tree object
 *
 *  Copyright (C) 2001 David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */
#include <config.h>

#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include <gtk/gtktreeview.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkcellrenderertext.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtksizegroup.h>
#include <gtk/gtkdialog.h>

#include <libgnome/gnome-i18n.h>

#include <libgnomevfs/gnome-vfs-mime-handlers.h>
#include <libgnomevfs/gnome-vfs-mime-utils.h>
#include <libgnomevfs/gnome-vfs-uri.h>

#include <glib/gmarkup.h>

#include "screem-tagtree.h"

#include "screem-window.h"
#include "screem-window-private.h"
#include "screem-editor.h"

#include "screem-file-browser.h"

#include "fileops.h"

#include "libegg/menu/egg-action-group.h"
#include "libegg/menu/egg-toggle-action.h"
#include "libegg/menu/egg-menu-merge.h"

#include "libegg/toolbar/eggtoolbar.h"

struct ScreemTagTreePrivate {
	GtkWidget *box;
	
	GtkWidget *view;

	GtkTreeStore *store;

	GHashTable *paths;

	ScreemWindow *window;
	ScreemFileBrowser *browser;

	GtkWidget *toolbar;
};

typedef struct {
	ScreemFileBrowser *browser;
	GtkTreeModel *model;
	GHashTable *paths;
} ScreemTagTreeBuildData;

typedef struct {
	ScreemTagTreeBuildData *data;
	GSList *states;
	GSList *iters;
} ScreemTagTreeParser;

static gboolean screem_tag_tree_parse_file( ScreemTagTreeBuildData *bdata, 
						const gchar *filename );
static gboolean screem_tag_tree_parse_string( ScreemTagTreeBuildData *bdata, 
						const gchar *str,
						guint len );

static void screem_tag_tree_start_element( GMarkupParseContext *context,
					const gchar *element,
					const gchar **attributes,
					const gchar **values,
					gpointer data,
					GError **error );
static void screem_tag_tree_end_element( GMarkupParseContext *context,
					const gchar *element,
					gpointer data,
					GError **error );
static void screem_tag_tree_text( GMarkupParseContext *context,
				const gchar *text,
				gsize len,
				gpointer data,
				GError **error );
static void screem_tag_tree_error( GMarkupParseContext *context,
				GError *error,
				gpointer data );
static void screem_tag_tree_parse_node( ScreemTagTreeParser *parser,
					const gchar **attributes,
					const gchar **values,
					gboolean group );
static void screem_tag_tree_parse_insert( ScreemTagTreeParser *parser,
					  const gchar **attributes,
					  const gchar **values );

static void screem_tag_tree_parse_param( ScreemTagTreeParser *parser,
					 const gchar **attributes,
					 const gchar **values );

static gboolean screem_tag_tree_clear_paths( gpointer key, gpointer value,
						gpointer data );


static void screem_tag_tree_icon_change( ScreemFileBrowser *browser, 
					const gchar *uri,
					const gchar *mime_type,
					GtkTreeIter *it,
					gpointer data );

static GMarkupParser parser_cbs = {
	screem_tag_tree_start_element,
	screem_tag_tree_end_element,
	screem_tag_tree_text,
	NULL,
	screem_tag_tree_error
};

typedef struct {
	gchar *name;
	gboolean required;
	gboolean vallist;
	gchar *type;
	gchar *defvalue;
	gchar *description;
} NodeAttribute;

typedef struct {
	gchar *name;
	
	gchar *open;
	gchar *close;
	
	gchar *file;
	gchar *text;
	
	gchar *description;

	gboolean isfunc;

	/* list of Attribute structs */
	GSList *attributes;
} NodeInfo;

typedef enum {
	PARSER_STATE_NONE = 0,
	PARSER_STATE_REF,
	PARSER_STATE_GROUP,
	PARSER_STATE_ENTRY,
	PARSER_STATE_FUNCTION,
	PARSER_STATE_INSERT,
	PARSER_STATE_DESCRIPTION,
	PARSER_STATE_PARAM,
	PARSER_STATE_ATTRIBUTE
} ScreemTagTreeParserState;

static const gchar *parser_tags[] = {
	"",
	"ref",
	"group",
	"entry",
	"function",
	"insert",
	"description",
	"param",
	"property",
};

static gboolean screem_tag_tree_parse_file( ScreemTagTreeBuildData *bdata, 
						const gchar *filename )
{
	GString *file;
	gboolean ret;
	
	ret = FALSE;
	file = load_file( filename );
	if( file ) {
		ret = screem_tag_tree_parse_string( bdata, file->str, file->len );
	}
	return ret;	
}

static gboolean screem_tag_tree_parse_string( ScreemTagTreeBuildData *bdata, 
						const gchar *str,
						guint len )
{
	GMarkupParseContext *ctx;
	ScreemTagTreeParser *data;
	gboolean ret;

	g_return_val_if_fail( bdata != NULL, FALSE );
	g_return_val_if_fail( str != NULL, FALSE );
	g_return_val_if_fail( len > 0, FALSE );

	ret = FALSE;

	data = g_new0( ScreemTagTreeParser, 1 );
	data->data = bdata;
	data->states = g_slist_prepend( data->states,
					GINT_TO_POINTER( PARSER_STATE_NONE ) );
	data->iters = g_slist_prepend( data->iters, NULL );
	ctx = g_markup_parse_context_new( &parser_cbs, 0, 
					  data, NULL );
	if( ctx ) {
		ret = g_markup_parse_context_parse( ctx, str, len, NULL );
		g_markup_parse_context_free( ctx );
	}
	if( data->iters->next ) {
		g_slist_foreach( data->iters,
				 (GFunc)gtk_tree_iter_free, NULL );
	}
	g_slist_free( data->iters );
	g_slist_free( data->states );
	g_free( data );

	return ret;	
}

static void screem_tag_tree_start_element( GMarkupParseContext *context,
					const gchar *element,
					const gchar **attributes,
					const gchar **values,
					gpointer data,
					GError **error )
{
	ScreemTagTreeParser *parser;
	ScreemTagTreeParserState state;

	parser = (ScreemTagTreeParser*)data;
	state = GPOINTER_TO_INT( parser->states->data );
	
	switch( state ) {
		case PARSER_STATE_NONE:
			if( ! strcmp( "ref", element ) ) {
				screem_tag_tree_parse_node( parser,
							    attributes,
							    values, TRUE );
				state = PARSER_STATE_REF;
			} else {
				/* unknown root element, ignore */
			}
			break;
		case PARSER_STATE_REF:
		/*ref can't have entry / function according to our dtd,
			but it simplifies adding the current doctype branch,
			so we allow it here */
		case PARSER_STATE_GROUP:
			if( ! strcmp( "group", element ) ) {
				screem_tag_tree_parse_node( parser,
							    attributes,
							    values, TRUE );
				state = PARSER_STATE_GROUP;
			} else if( ! strcmp( "entry", element ) ) {
				screem_tag_tree_parse_node( parser,
							    attributes,
							    values, FALSE );
				state = PARSER_STATE_ENTRY;
			} else if( ! strcmp( "function", element ) ) {
				screem_tag_tree_parse_node( parser,
							    attributes,
							    values, FALSE );
				state = PARSER_STATE_FUNCTION;
			}
			break;
		case PARSER_STATE_ENTRY:
			if( ! strcmp( "insert", element ) ) {
				screem_tag_tree_parse_insert( parser,
								attributes,
								values );
				state = PARSER_STATE_INSERT;
			} else if( ! strcmp( "description", element ) ) {
				state = PARSER_STATE_DESCRIPTION;
			} else if( ! strcmp( "property", element ) ) {
				screem_tag_tree_parse_param( parser,
				 			attributes,
							values );
				state = PARSER_STATE_ATTRIBUTE;
			}
			break;
		case PARSER_STATE_FUNCTION:
			if( ! strcmp( "insert", element ) ) {
				screem_tag_tree_parse_insert( parser,
								attributes,
								values );
				state = PARSER_STATE_INSERT;
			} else if( ! strcmp( "description", element ) ) {
				state = PARSER_STATE_DESCRIPTION;
			} else if( ! strcmp( "param", element ) ) {
				screem_tag_tree_parse_param( parser,
				 			attributes,
							values );
				state = PARSER_STATE_PARAM;
			} else {
				state = PARSER_STATE_NONE;
			}
			break;
		case PARSER_STATE_INSERT:
		case PARSER_STATE_DESCRIPTION:
		case PARSER_STATE_PARAM:
		case PARSER_STATE_ATTRIBUTE:
			/* states should not occur */
			state = PARSER_STATE_NONE;
			break;
		default:
			/* unknown state, shouldn't happen */
			state = PARSER_STATE_NONE;
			break;	
	}

	if( state != PARSER_STATE_NONE ) {
		parser->states = g_slist_prepend( parser->states,
						  GINT_TO_POINTER( state ) );
	}
}

static void screem_tag_tree_end_element( GMarkupParseContext *context,
					const gchar *element,
					gpointer data,
					GError **error )
{
	ScreemTagTreeParser *parser;
	ScreemTagTreeParserState state;
	ScreemTagTreeBuildData *bdata;
	GSList *top;
	NodeInfo *info;
	GValue value = { 0 };
	
	parser = (ScreemTagTreeParser*)data;
	bdata = parser->data;
	state = GPOINTER_TO_INT( parser->states->data );
	top = parser->states;

	switch( state ) {
		case PARSER_STATE_NONE:
			/* shouldn't happen */
			state = PARSER_STATE_NONE;
			break;
		default:
			if( strcmp( parser_tags[ state ], element ) ) {
				state = PARSER_STATE_NONE;
			}
			break;
	}

	if( state != PARSER_STATE_NONE ) {
		parser->states = top->next;
		top->next = NULL;
		g_slist_free( top );
		if( state == PARSER_STATE_REF ||
		    state == PARSER_STATE_GROUP ||
		    state == PARSER_STATE_ENTRY ||
		    state == PARSER_STATE_FUNCTION ) {
			top = parser->iters;
			parser->iters = top->next;
			top->next = NULL;
			if( top->data ) {
				info = NULL;
				gtk_tree_model_get_value( GTK_TREE_MODEL( bdata->model ),
						  top->data, FILE_BROWSER_DATA_COL,
						  &value );
				info = g_value_get_pointer( &value );
				g_value_unset( &value );
				if( info && info->attributes ) {
					info->attributes = g_slist_reverse( info->attributes );
				}
	
 				gtk_tree_iter_free( (GtkTreeIter*)top->data );
			}
			g_slist_free( top );
		}
	}
}

static void screem_tag_tree_text( GMarkupParseContext *context,
				const gchar *text,
				gsize len,
				gpointer data,
				GError **error )
{
	ScreemTagTreeParser *parser;
	ScreemTagTreeParserState state;
	ScreemTagTreeBuildData *bdata;
	GSList *top;
	
	GtkTreeIter *it;
	GValue value = { 0 };
	NodeInfo *info;
	
	parser = (ScreemTagTreeParser*)data;
	state = GPOINTER_TO_INT( parser->states->data );
	bdata = parser->data;
		
	top = parser->iters;
	
	it = top->data;
      	gdk_threads_enter();
	info = NULL;
	if( it ) {
		gtk_tree_model_get_value( GTK_TREE_MODEL( bdata->model ),
					  it, FILE_BROWSER_DATA_COL,
					  &value );
		info = g_value_get_pointer( &value );
		g_value_unset( &value );
	}	
     	gdk_threads_leave();
	
	switch( state ) {
		case PARSER_STATE_INSERT:
			if( info ) {
				info->text = g_strdup( text );
				info->text = g_strstrip( info->text );
			}
			break;
		case PARSER_STATE_DESCRIPTION:
			if( info ) {
				info->description = g_strdup( text );
				info->description = g_strstrip( info->description );
			}
			break;
		case PARSER_STATE_PARAM:
		case PARSER_STATE_ATTRIBUTE:
			if( info && info->attributes ) {
				NodeAttribute *attr;

				attr = (NodeAttribute*)info->attributes->data;

				if( attr ) {
					attr->description = g_strdup( text );
					attr->description = g_strstrip( attr->description );
				}
			}
			break;
		default:
			/* shouldn't happen */
			break;
	}
}

static void screem_tag_tree_error( GMarkupParseContext *context,
				GError *error,
				gpointer data )
{
	
}

static void screem_tag_tree_parse_node( ScreemTagTreeParser *parser,
					const gchar **attributes,
					const gchar **values,
					gboolean group )
{
	GSList *top;
	GtkTreeIter it;
	GtkTreeIter *iter;
	const gchar *name;
	const gchar *description;
	gint i;
	GtkTreePath *path;
	NodeInfo *info;

	GString *fullpath;
	GSList *paths;
	
	ScreemTagTreeBuildData *bdata;
	
	bdata = parser->data;
	
	top = parser->iters;
	
	info = NULL;	
	name = description = NULL;
	for( i = 0; attributes[ i ]; ++ i ) {
		if( ! strcmp( "name", attributes[ i ] ) ) {
			name = values[ i ];
		} else if( ! strcmp( "description", attributes[ i ] ) ) {
			description = values[ i ];
		}
	}
	
	/* name="" is required, but we will create one if it doesn't exist */
	if( ! name ) {
		name = _( "Unknown Resource" );
	}
	
	/* see if we already have a node matching name at this level,
	   if so we just push that iter instead of creating a new one */

      	gdk_threads_enter();
	fullpath = g_string_new( name );
	for( paths = top; paths; paths = paths->next ) {
		GValue value = { 0 };
		GtkTreeIter *pit;
		
		pit = (GtkTreeIter*)paths->data;
		if( pit ) {
			gtk_tree_model_get_value( GTK_TREE_MODEL( bdata->model ),
						  pit, FILE_BROWSER_DATA_COL,
						  &value );
			info = g_value_get_pointer( &value );
			g_value_unset( &value );
			g_string_prepend_c( fullpath, '/' );
			g_string_prepend( fullpath, info->name );
		}
	}
	g_string_prepend_c( fullpath, '/' );
	path = g_hash_table_lookup( bdata->paths, fullpath->str );

	if( path && ! gtk_tree_model_get_iter( GTK_TREE_MODEL( bdata->model ),
						&it, path ) ) {
		path = NULL;
	} else if( path && ! strcmp( _( "Current Doctype" ), fullpath->str + 1 ) ) {
		gtk_tree_store_remove( GTK_TREE_STORE( bdata->model ), &it );
		g_hash_table_foreach_steal( bdata->paths,
				screem_tag_tree_clear_paths,
				_( "Current Doctype" ) );
		path = NULL;
	}
	
	if( ! path ) {
		GdkPixbuf *pixbuf;
		
		gtk_tree_store_append( GTK_TREE_STORE( bdata->model ),
					&it, (GtkTreeIter*)top->data );						 
		
		pixbuf = NULL;
		if( group ) {
			pixbuf = screem_file_browser_get_icon( bdata->browser,
								"/", 16, 16,
								FALSE );
		}
		if( pixbuf ) {
			gtk_tree_store_set( GTK_TREE_STORE( bdata->model ),
					   &it, FILE_BROWSER_ICON_COL,
					   pixbuf, -1 );
			g_object_unref( pixbuf );
		}
		gtk_tree_store_set( GTK_TREE_STORE( bdata->model ), 
					&it, FILE_BROWSER_NAME_COL,
					name, -1 );
		info = g_new0( NodeInfo, 1 );
		info->name = g_strdup( name );
		if( description && group ) {
			info->description = g_strdup( description );
		}
		gtk_tree_store_set( GTK_TREE_STORE( bdata->model ),
					&it, FILE_BROWSER_DATA_COL,
					info, -1 );

		path = NULL;
	 }
	
	iter = gtk_tree_iter_copy( &it );
	parser->iters = g_slist_prepend( parser->iters, iter );
	
	if( ! path ) {
		path = gtk_tree_model_get_path( GTK_TREE_MODEL( bdata->model ),
						&it );
		g_hash_table_insert( bdata->paths, fullpath->str,
				     path );
		path = NULL;
	}
	
	g_string_free( fullpath, ( path != NULL ) );
	
     	gdk_threads_leave();
}


static void screem_tag_tree_parse_insert( ScreemTagTreeParser *parser,
					const gchar **attributes,
					const gchar **values )
{
	ScreemTagTreeParserState state;
	GSList *top;
	
	GtkTreeIter *it;
	GValue value = { 0 };
	NodeInfo *info;
	gint i;
	ScreemTagTreeBuildData *bdata;
	
	bdata = parser->data;
	state = GPOINTER_TO_INT( parser->states->data );
	top = parser->iters;

      	gdk_threads_enter();
	
	it = top->data;
	info = NULL;
	if( it ) {
		gtk_tree_model_get_value( GTK_TREE_MODEL( bdata->model ),
					  it, FILE_BROWSER_DATA_COL,
					  &value );
		info = g_value_get_pointer( &value );
		g_value_unset( &value );
	}
	if( info ) {
		info->isfunc = ( state == PARSER_STATE_FUNCTION );
		for( i = 0; attributes[ i ]; ++ i ) {
			const gchar *attr;
			const gchar *val;
		
			attr = attributes[ i ];
			val = values[ i ];
		
			if( ! strcmp( "href", attr ) ) {
				GdkPixbuf *pixbuf;	
				gchar *path;
				path = relative_to_full( val, 
							 RESOURCE_PATH );
				info->file = path;
				pixbuf = screem_file_browser_get_icon( bdata->browser,
									info->file,
									16, 16, FALSE );
				gtk_tree_store_set( GTK_TREE_STORE( bdata->model ), 
						    it, FILE_BROWSER_ICON_COL,
						    pixbuf, -1 );
				g_object_unref( pixbuf );
			} else if( ! strcmp( "open", attr ) ) {
				info->open = g_strdup( val );
			} else if( ! strcmp( "close", attr ) ) {
				info->close = g_strdup( val );
			}
		}
	}
      	gdk_threads_leave();
}


static void screem_tag_tree_parse_param( ScreemTagTreeParser *parser,
					 const gchar **attributes,
					 const gchar **values )
{
	NodeAttribute *attr;
	gint i;
	gint j;
	const gchar *attribute;
	const gchar *value;
	GSList *top;
	GtkTreeIter *it;
	ScreemTagTreeBuildData *bdata;
	GValue val = { 0 };
	NodeInfo *info;


	top = parser->iters;
	g_assert( top );
	bdata = parser->data;
	
	it = top->data;
	gdk_threads_enter();
	info = NULL;
	if( it ) {
		gtk_tree_model_get_value( GTK_TREE_MODEL( bdata->model ),
					it, FILE_BROWSER_DATA_COL,
					&val );
		info = g_value_get_pointer( &val );
		g_value_unset( &val );
	}
	gdk_threads_leave();

	if( ! info ) {
		g_warning( "<param> with no current node\n" );
		return;
	}
	
	attr = g_new0( NodeAttribute, 1 );
	
	for( i = 0, j = 0; attributes[ i ]; ++ i ) {
		attribute = attributes[ i ];
		
		value = NULL;
		if( values[ j ] ) {
			value = values[ j ];
			++ j;
		}

		if( value ) {
			if( ! strcmp( "name", attribute ) ) {
				attr->name = g_strdup( value );
				attr->name = g_strstrip( attr->name );
			} else if( ! strcmp( "required", attribute ) ) {
				attr->required = ( *value == '1');
			} else if( ! strcmp( "vallist", attribute ) ) {
				attr->vallist = ( *value == '1' );
			} else if( ! strcmp( "type", attribute ) ) {
				attr->type = g_strdup( value );
			} else if( ! strcmp( "default", attribute ) ) {
				attr->defvalue = g_strdup( value );
				attr->defvalue = g_strstrip( attr->defvalue );
			}
		}
	}
	info->attributes = g_slist_prepend( info->attributes, attr );
}

enum {
	ARG_0,
	ARG_WINDOW
};
static gint screem_tag_tree_compare_func( GtkTreeModel *model, 
				    	  GtkTreeIter *a, GtkTreeIter *b,
				    	  gpointer data );
static void screem_tagtree_sel_change( GtkTreeSelection *selection,
					ScreemTagTree *tree );
static ScreemFileBrowser *screem_tag_tree_get_browser( void );
static void screem_tag_tree_set_model( ScreemTagTree *tree, 
					GtkTreeModel *model );
static gpointer screem_tag_tree_create( gpointer data );

static void screem_tag_tree_class_init( ScreemTagTreeClass *klass );
static void screem_tag_tree_init( ScreemTagTree *tag_tree );
static void screem_tag_tree_finalize( GObject *object );
static void screem_tag_tree_set_prop( GObject *object, guint property_id,
				    const GValue *value, GParamSpec *pspec );
static void screem_tag_tree_get_prop( GObject *object, guint property_id,
				      GValue *value, GParamSpec *pspec);
static void screem_tag_tree_size_request( GtkWidget *widget,
                           	  	  GtkRequisition *requisition );
static void screem_tag_tree_size_allocate( GtkWidget *widget,
                              		   GtkAllocation *allocation );

ScreemTagTree *screem_tag_tree_new()
{
	ScreemTagTree *tree;
	GType type;

	type = screem_tag_tree_get_type();
	
	tree = SCREEM_TAG_TREE( g_object_new( type, NULL ) );

	return tree;
}

gboolean screem_tag_tree_load_trees( void )
{
	ScreemFileBrowser *browser;
	static GThread *thread = NULL;
	
	browser = screem_tag_tree_get_browser();
	
	/* spawn load thread, so the ui will still be interactive
	   as tag tree can take a while to load */
	if( ! thread ) {
		thread = g_thread_create( screem_tag_tree_create, browser,
					  FALSE, NULL );
	} else {
		g_warning( "screem_tag_tree_load_trees() called more than once\n" );
	}
	
	return FALSE;
}

void screem_tag_tree_add_dtd( ScreemTagTree *tag_tree, ScreemDTD *dtd )
{
	GString *str;
	const GSList *elements;
	ScreemTagTreeBuildData *bdata;
	
	if( ! dtd ) {
		return;
	}
		
	elements = screem_dtd_get_elements( dtd );
	
	str = g_string_new( "<?xml version=\"1.0\"?>" );
	g_string_append_printf( str, "<ref name=\"%s\" description=\"%s\">",
				_( "Current Doctype" ),
				_( "Elements for the current doctype" ) );
	while( elements ) {
		const gchar *name;
		const gchar *desc;
		ScreemDTDClose close;
		const GSList *attrs;
		const gchar *adesc;

		name = screem_dtd_element_get_name( (ScreemDTDElement*)elements->data );
		desc = screem_dtd_element_get_description( (ScreemDTDElement*)elements->data );		
		close = screem_dtd_element_close_state( dtd, name );
		attrs = screem_dtd_element_get_attrs( (ScreemDTDElement*)elements->data );
		
		g_string_append_printf(str, 
				"<entry name=\"%s\"><insert open=\"&lt;%s&gt;\"",
				name, name );
		if( close != MUSTNT ) {
			g_string_append_printf( str, " close=\"&lt;/%s&gt;\"",
						name );
		}
		g_string_append( str, "/>");
		if( desc ) {
			g_string_append_printf( str, 
					"<description>%s</description>",
					desc );
		}

		while( attrs ) {
			const ScreemDTDAttribute *attr;

			attr = (const ScreemDTDAttribute*)attrs->data;
			adesc = screem_dtd_attribute_get_description( attr );
			if( ! adesc ) {
				adesc = "";
			}
			g_string_append_printf( str,
					"<property name=\"%s\" required=\"%i\" vallist=\"0\" default=\"%s\">%s</property>",
					screem_dtd_attribute_get_name( attr ),
					screem_dtd_attribute_required( attr ),
					"",
					adesc
					);
			attrs = attrs->next;
		}
		
		g_string_append( str, "</entry>" );
	
		elements = elements->next;
	}

	g_string_append( str, "</ref>" );

	bdata = g_new0( ScreemTagTreeBuildData, 1 );
	bdata->browser = tag_tree->private->browser;
	bdata->paths = tag_tree->private->paths;
	bdata->model = GTK_TREE_MODEL( tag_tree->private->store );

	screem_tag_tree_parse_string( bdata, str->str, str->len );
	g_string_free( str, TRUE );
	g_free( bdata );
}


/* static stuff */
static gboolean screem_tag_tree_click( GtkTreeView *tree_view,
				       GtkTreePath *path,
				       GtkTreeViewColumn *column,
				       gpointer data )
{
	GtkTreeIter iter;
	GtkTreeModel *model;
	GtkTreeSelection *sel;

	ScreemTagTree *tag_tree;

	tag_tree = SCREEM_TAG_TREE( data );


	model = GTK_TREE_MODEL( tag_tree->private->store );
	sel = gtk_tree_view_get_selection( GTK_TREE_VIEW( tree_view ) );

	if( sel ) {
		GValue value = { 0 };
		NodeInfo *info;
		ScreemWindow *window;
		ScreemEditor *editor;

		window = SCREEM_WINDOW( tag_tree->private->window );

		editor = SCREEM_EDITOR( window->details->editor );

		gtk_tree_selection_get_selected( sel, &model, &iter );
		gtk_tree_model_get_value( model, &iter, FILE_BROWSER_DATA_COL,
					 &value );

		info = g_value_get_pointer( &value );

		if( info && ( ! info->text ) && info->file ) {
			GnomeVFSURI *vuri;
			
			vuri = gnome_vfs_uri_new( info->file );
			if( ! vuri ) {
				screem_window_show_error( window, _( "Invalid URI for requested resource" ) );
			} else if( gnome_vfs_uri_is_local( vuri ) ||
				  ! window->details->offline ) {
				GString *data;
				
				data = load_file( info->file );
				/* now set info->text to be the file contents,
			   	this way we don't refetch files we already have */
				info->text = data->str;
				g_string_free( data, FALSE );
			} else {
				screem_window_show_message( window, _( "Remote resources are not available in offline mode" ) );	
			}

			if( vuri ) {	
				gnome_vfs_uri_unref( vuri );
			}
		} else if( info && info->open ) {
			screem_editor_insert_markup( editor, info->open,
						     info->close );
		}
		/* we do info->text here so you can have something like
		 * <insert open="&lt;title&gt;" close="&lt;title&gt;">
		 *      page title
		 * </insert>
		 */
		if( info && info->text ) {
			screem_editor_insert( editor, -1, info->text );
		}
		g_value_unset( &value );
		
	} else {
		/* hmm, no iterator */
		
	}

	return (sel != NULL);
}


static gboolean screem_tag_tree_press( GtkWidget *widget, 
				       GdkEventButton *event,
				       ScreemTagTree *tag_tree )
{
	
	return FALSE;
}

static gint screem_tag_tree_compare_func( GtkTreeModel *model, 
				    	  GtkTreeIter *a, GtkTreeIter *b,
				    	  gpointer data )
{
	GValue avalue = { 0 };
	GValue bvalue = { 0 };
        NodeInfo *ainfo;
        NodeInfo *binfo;

	int ret = 0;

	gtk_tree_model_get_value( model, a, FILE_BROWSER_DATA_COL, &avalue );
	gtk_tree_model_get_value( model, b, FILE_BROWSER_DATA_COL, &bvalue );

	ainfo = g_value_get_pointer( &avalue );
	binfo = g_value_get_pointer( &bvalue );
	
	if( ainfo && ! binfo ) {
		ret = 1;
	} else if( binfo && !ainfo ) {
		ret = -1;
	} else if( ! ainfo && ! binfo ) {
		ret = 0;
	} else if( ! strcmp( _( "Current Doctype" ), ainfo->name ) ) {
		ret = -1;
	} else if( ! strcmp( _( "Current Doctype" ), binfo->name ) ) {
		ret = 1;
	} else {
		ret = strcmp( ainfo->name, binfo->name );
	}
	
	g_value_unset( &avalue );
	g_value_unset( &bvalue );

	return ret;
}

static void screem_tagtree_sel_change( GtkTreeSelection *selection,
					ScreemTagTree *tree )
{
	ScreemTagTreePrivate *priv;
	ScreemWindow *window;
	gint rows;
	GtkTreeModel *model;
	GtkTreeIter iter;
	GValue value = { 0 };
	NodeInfo *info;
	
	priv = tree->private;
	
	window = SCREEM_WINDOW( priv->window );
	rows = gtk_tree_selection_count_selected_rows( selection );
	if( rows > 0 ) {
		gtk_tree_selection_get_selected( selection, &model, &iter );
		gtk_tree_model_get_value( model, &iter, FILE_BROWSER_DATA_COL,
					 &value );
		info = g_value_get_pointer( &value );
		if( info && info->description ) {
			screem_window_show_message( window, info->description );
		}
		g_value_unset( &value );
	}
	gtk_widget_set_sensitive( priv->toolbar, ( rows > 0 ) );
}

static ScreemFileBrowser *screem_tag_tree_get_browser( void )
{
	static ScreemFileBrowser *browser = NULL;
	
	if( ! browser ) {
		GHashTable *table;
		
		browser = screem_file_browser_new();
		screem_file_browser_set_mode( browser,
						FILE_BROWSE_RECURSE );
		screem_file_browser_set_sort_func( browser,
					   screem_tag_tree_compare_func,
					   NULL );
	
		g_signal_connect( G_OBJECT( browser ),
				  "icon_change", 
				  G_CALLBACK( screem_tag_tree_icon_change ),
				  NULL );

		table = g_hash_table_new( g_str_hash, g_str_equal );
		g_object_set_data( G_OBJECT( browser ), "paths", table );
	}
	
	return browser;
}

static void screem_tag_tree_set_model( ScreemTagTree *tree, GtkTreeModel *model )
{
	ScreemTagTreePrivate *priv;
	
	g_return_if_fail( SCREEM_IS_TAG_TREE( tree ) );
	g_return_if_fail( model != NULL );
	
	priv = tree->private;

	gtk_tree_view_set_model( GTK_TREE_VIEW( tree->private->view ), model );
	
	if( GTK_IS_TREE_SORTABLE( model ) ) {
		GtkTreeModelSort *sortable;
		
		sortable = GTK_TREE_MODEL_SORT( model );
		model = gtk_tree_model_sort_get_model( sortable );
	}
	priv->store = GTK_TREE_STORE( model );	
}

static gpointer screem_tag_tree_create( gpointer data )
{
	gchar *mime_type;
	gchar *file;
	gint num;
	gint pos;
	struct dirent **namelist;

	gint i;
	const gchar *treepaths[] = {
		DATADIR"/screem/tagtrees",
		NULL,
		NULL
	};
	ScreemTagTreeBuildData *bdata;
		
	bdata = g_new0( ScreemTagTreeBuildData, 1 );
	bdata->browser = SCREEM_FILE_BROWSER( data );
	bdata->paths = g_object_get_data( G_OBJECT( data ), "paths" );
	bdata->model = screem_file_browser_get_model(SCREEM_FILE_BROWSER(data));
	bdata->model = gtk_tree_model_sort_get_model( GTK_TREE_MODEL_SORT( bdata->model ) );
	
	/* load tag trees */
	treepaths[ 1 ] = screem_get_dot_dir();

	for( i = 0; treepaths[ i ]; ++ i ) {
		num = scandir( treepaths[ i ], &namelist, 0, alphasort );
		
		for( pos = 0; pos < num; pos ++ ) {
			gchar *path;
			
			file = namelist[ pos ]->d_name;
			path = g_strconcat( treepaths[ i ],
						G_DIR_SEPARATOR_S,
					    	file, NULL );
			mime_type = gnome_vfs_get_mime_type( path );
			if( mime_type &&
		    	! strcmp( "application/x-screem-tag-tree", mime_type ) ) {
				screem_tag_tree_parse_file( bdata, path );
			}
			if( mime_type ) {
				g_free( mime_type );
			}
			g_free( path );
			free( namelist[ pos ] );
		}
		if( num > 0 ) {
			free( namelist );
		}
	}
	g_free( (gchar*)treepaths[ 1 ] );

	return NULL;
}

static gboolean screem_tag_tree_clear_paths( gpointer key, gpointer value,
						gpointer data )
{
	const gchar *prefix;
	gboolean ret;

	prefix = (const gchar*)data;
	
	ret = FALSE;
	if( ! strncmp( prefix, key + 1, strlen( prefix ) ) ) {
		g_free( key );
		gtk_tree_path_free( value );
		ret = TRUE;
	}

	return ret;
}

static void add_widget( EggMenuMerge *merge, GtkWidget *widget, ScreemTagTree *tree )
{
	ScreemTagTreePrivate *priv;
	GtkWidget *box;
	
	priv = tree->private;
	box = priv->box;
	
	if( EGG_IS_TOOLBAR( widget ) ) {
		egg_toolbar_set_show_arrow( EGG_TOOLBAR( widget ), TRUE );
		egg_toolbar_set_tooltips( EGG_TOOLBAR( widget ), TRUE );
		
		gtk_box_pack_start( GTK_BOX( box ), widget,
				    FALSE, TRUE, 0 );
		gtk_widget_set_sensitive( widget, FALSE );
		priv->toolbar = widget;
	}
}

static void screem_tag_tree_tag_help( EggAction *action, ScreemTagTree *tree )
{
	ScreemTagTreePrivate *priv;
	GtkTreeView *view;
	GtkTreeSelection *selection;
	GtkTreeModel *model;
	GtkTreeIter it;
	gint rows;
	GValue value = { 0 };
	NodeInfo *info;
	
	priv = tree->private;

	view = GTK_TREE_VIEW( priv->view );
	selection = gtk_tree_view_get_selection( view );
	
	rows = gtk_tree_selection_count_selected_rows( selection );
	if( rows > 0 ) {
		gtk_tree_selection_get_selected( selection, &model, &it );
		gtk_tree_model_get_value( model, &it, FILE_BROWSER_DATA_COL,
					 &value );
		info = g_value_get_pointer( &value );
		if( info ) {
			GtkWidget *dialog;
			GtkWidget *box;
			GtkWidget *widget;
			GtkWidget *notebook;
			gchar *title;
			ScreemWindow *window;
			ScreemEditor *editor;
			GtkWidget *sw;
			GtkTextBuffer *buffer;
			gchar *txt;
			GtkSizeGroup *group;
			GSList *attrs;
			NodeAttribute *attr;
			GtkWidget *attrbox;
			gchar *attrtxt;
			
			g_object_get( G_OBJECT( tree ), "window", &window, NULL );
			editor = SCREEM_EDITOR( window->details->editor );
			
			box = gtk_vbox_new( FALSE, 0 );
			gtk_box_set_spacing( GTK_BOX( box ), 6 );
			gtk_container_set_border_width( GTK_CONTAINER( box ), 12 );
		
			if( ! info->isfunc ) {
				title = g_strconcat( "<span size=\"xx-large\">",
							info->name,
							"</span>",
							NULL );
			} else {
				title = g_strconcat( "<span size=\"xx-large\">",
							info->name,
							"()</span>",
							NULL );
			}
			widget = gtk_label_new( "" );
			gtk_label_set_use_markup( GTK_LABEL( widget ), TRUE );
			gtk_label_set_markup( GTK_LABEL( widget ), title );
			g_free( title );
			gtk_label_set_justify( GTK_LABEL( widget ), GTK_JUSTIFY_LEFT );
			gtk_misc_set_alignment( GTK_MISC( widget ), 0.0, 0.0 );
			gtk_box_pack_start( GTK_BOX( box ), widget, FALSE, TRUE, 0 );
	

			widget = gtk_label_new( info->description );
			gtk_label_set_justify( GTK_LABEL( widget ), GTK_JUSTIFY_LEFT );
			gtk_misc_set_alignment( GTK_MISC( widget ), 0.0, 0.0 );
			gtk_label_set_line_wrap( GTK_LABEL( widget ), TRUE );
			gtk_box_pack_start( GTK_BOX( box ), widget, FALSE, TRUE, 0 );

			txt = NULL;
			if( ( ! info->text ) && info->file ) {
				GnomeVFSURI *vuri;
			
				vuri = gnome_vfs_uri_new( info->file );
				if( ! vuri ) {
					txt = g_strdup( _( "Invalid URI for requested resource" ) );
				} else if( gnome_vfs_uri_is_local( vuri ) ||
					  ! window->details->offline ) {
					GString *data;
				
					data = load_file( info->file );
					/* now set info->text to be the file contents,
				   	this way we don't refetch files we already have */
					info->text = data->str;
					g_string_free( data, FALSE );
				} else {
					txt = g_strdup( _( "Remote resources are not available in offline mode" ) );	
				}

				if( vuri ) {	
					gnome_vfs_uri_unref( vuri );
				}
			} 
			if( info->open ) {
				txt = g_strconcat( info->open, info->close, NULL );
			}
			if( info->text ) {
				g_free( txt );
				txt = g_strdup( info->text );
			}
			
			notebook = gtk_notebook_new();
			gtk_notebook_set_show_border( GTK_NOTEBOOK( notebook ),
							FALSE );
			gtk_notebook_set_tab_pos( GTK_NOTEBOOK( notebook ),
						GTK_POS_BOTTOM );
			gtk_box_pack_start( GTK_BOX( box ), notebook, TRUE, TRUE, 0 );
	

			if( info->attributes ) {
				sw = gtk_scrolled_window_new( NULL, NULL );
				gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( sw ),
								GTK_POLICY_NEVER,
								GTK_POLICY_AUTOMATIC );
				gtk_scrolled_window_set_shadow_type( GTK_SCROLLED_WINDOW( sw ),
								GTK_SHADOW_IN );

				group = gtk_size_group_new( GTK_SIZE_GROUP_HORIZONTAL );

				attrbox = gtk_vbox_new( FALSE, 0 );

				for( attrs = info->attributes; attrs; attrs = attrs->next ) {
					attr = (NodeAttribute*)attrs->data;

					if( attr->type ) {
						attrtxt = g_strconcat( "<b><i>", attr->name, "(",
									attr->type,
									") - </i></b>",
									attr->description,
									NULL );
					} else {
						attrtxt = g_strconcat( "<b><i>", attr->name,
									" - </i></b>",
									attr->description,
									NULL );
					}
					widget = gtk_label_new( "" );
					gtk_label_set_markup( GTK_LABEL( widget ), attrtxt );
					gtk_label_set_use_markup( GTK_LABEL( widget ), TRUE );
					g_free( attrtxt );
					gtk_size_group_add_widget( group, widget );
				
					gtk_label_set_justify( GTK_LABEL( widget ), GTK_JUSTIFY_LEFT );
					gtk_misc_set_alignment( GTK_MISC( widget ), 0.0, 0.0 );

					gtk_box_pack_start( GTK_BOX( attrbox ), widget, FALSE, TRUE, 0 );
				}
				gtk_container_set_border_width( GTK_CONTAINER( attrbox ), 6 );
				gtk_box_set_spacing( GTK_BOX( attrbox ), 6 );
				gtk_scrolled_window_add_with_viewport( GTK_SCROLLED_WINDOW( sw ),
								attrbox );
				
				if( info->isfunc ) {
					widget = gtk_label_new( _( "Parameters" ) );
				} else {
					widget = gtk_label_new( _( "Attributes" ) );
				}
				
				gtk_notebook_append_page( GTK_NOTEBOOK( notebook ),
							  sw, widget );
			}
			
			if( txt ) {
				sw = gtk_scrolled_window_new( NULL, NULL );
				gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( sw ),
								GTK_POLICY_AUTOMATIC,
								GTK_POLICY_AUTOMATIC );
				gtk_scrolled_window_set_shadow_type( GTK_SCROLLED_WINDOW( sw ),
								GTK_SHADOW_IN );
				widget = gtk_source_view_new();
				gtk_text_view_set_editable( GTK_TEXT_VIEW( widget ), FALSE );
				gtk_text_view_set_wrap_mode( GTK_TEXT_VIEW( widget ), GTK_WRAP_WORD );
				buffer = gtk_text_view_get_buffer( GTK_TEXT_VIEW( widget ) );
				gtk_text_buffer_set_text( buffer, txt, -1 );
				gtk_container_add( GTK_CONTAINER( sw ), widget );

				widget = gtk_label_new( _( "Text to Insert" ) );
				gtk_notebook_append_page( GTK_NOTEBOOK( notebook ),
							 sw, widget );
				dialog = gtk_dialog_new_with_buttons( _( "Tag Help - Screem" ),
								GTK_WINDOW( window ),
								GTK_DIALOG_MODAL | 
								GTK_DIALOG_NO_SEPARATOR,
								GTK_STOCK_CLOSE,
								GTK_RESPONSE_CLOSE,
								GTK_STOCK_ADD,
								GTK_RESPONSE_APPLY,
								NULL);
			} else {
				dialog = gtk_dialog_new_with_buttons( _( "Tag Help - Screem" ),
								GTK_WINDOW( window ),
								GTK_DIALOG_MODAL | 
								GTK_DIALOG_NO_SEPARATOR,
								GTK_STOCK_CLOSE,
								GTK_RESPONSE_CLOSE,
								NULL );
			}
			gtk_widget_show_all( box );
			
			gtk_box_pack_start( GTK_BOX( GTK_DIALOG( dialog )->vbox ), box, TRUE, TRUE, 0 );

			if( gtk_dialog_run( GTK_DIALOG( dialog ) ) == GTK_RESPONSE_APPLY ) {
				/* insert txt into the editor */
				if( info->open ) {
					screem_editor_insert_markup( editor, info->open, info->close );
				} else {
					screem_editor_insert( editor, -1, txt );
				}
			}
			g_free( txt );
			
			
			gtk_widget_destroy( dialog );
		}
		g_value_unset( &value );
	} else {
		g_warning( "no item selected for info\n" );
	}
}

static void screem_tag_tree_icon_change( ScreemFileBrowser *browser, 
					const gchar *uri,
					const gchar *mime_type,
					GtkTreeIter *it,
					gpointer data )
{
	GtkTreeModel *model;
	NodeInfo *info;
	GdkPixbuf *pixbuf;
	
	model = screem_file_browser_get_model( browser );
	model = gtk_tree_model_sort_get_model( GTK_TREE_MODEL_SORT( model ) );

	
	gtk_tree_model_get( model, it, 
			FILE_BROWSER_DATA_COL, &info,
			-1 );

	if( info ) {
		pixbuf = NULL;
		if( info->file ) {
			pixbuf = screem_file_browser_get_icon( browser,
								info->file,
								16, 16, FALSE );
		} else if( gtk_tree_model_iter_has_child( model, it ) ) {
			pixbuf = screem_file_browser_get_icon( browser,
								"/",
								16, 16, FALSE );
		}
		gtk_tree_store_set( GTK_TREE_STORE( model ), 
				it, FILE_BROWSER_ICON_COL,
				pixbuf, -1 );
		if( pixbuf ) {
			g_object_unref( pixbuf );
		}
	}
}

/* G Object stuff */
#define PARENT_TYPE GTK_TYPE_BIN

static gpointer parent_class;

static void screem_tag_tree_class_init( ScreemTagTreeClass *klass )
{
	GObjectClass *object_class;
	GtkWidgetClass *widget_class;

	GParamSpec *pspec;

	object_class = G_OBJECT_CLASS( klass );
	widget_class = (GtkWidgetClass *)klass;

	object_class->finalize = screem_tag_tree_finalize;
	parent_class = g_type_class_peek_parent( klass );

	object_class->get_property = screem_tag_tree_get_prop;
	object_class->set_property = screem_tag_tree_set_prop;

	pspec = g_param_spec_pointer( "window", "window",
				      "window object",
				      G_PARAM_READABLE |
				      G_PARAM_WRITABLE );
	g_object_class_install_property( object_class,
					 ARG_WINDOW, pspec );
	
	widget_class->size_request = screem_tag_tree_size_request;
	widget_class->size_allocate = screem_tag_tree_size_allocate;
}

static void screem_tag_tree_init( ScreemTagTree *tag_tree )
{
	GtkWidget *sw;
	GtkCellRenderer *rend;
	GtkCellRenderer *prend;
	GtkTreeViewColumn *col;
	GtkTreeSelection *selection;

	static EggActionGroupEntry action_entries[] = {
		{ 
		   "Tag Help", "Tag Help",
		   GTK_STOCK_DIALOG_INFO, NULL, "Help",
	  	   G_CALLBACK( screem_tag_tree_tag_help ), NULL 
		},
	};
	static guint action_n_entries=G_N_ELEMENTS( action_entries );
	EggActionGroup *group;
	EggMenuMerge *merge;
	gint i;
	
	ScreemTagTreePrivate *priv;
	
	priv = tag_tree->private = g_new0( ScreemTagTreePrivate, 1 );

	priv->browser = screem_tag_tree_get_browser();
		
	priv->box = gtk_vbox_new( FALSE, 0 );
	gtk_container_add( GTK_CONTAINER( tag_tree ), priv->box );
		
	sw = gtk_scrolled_window_new( NULL, NULL );
	gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( sw ),
					GTK_POLICY_AUTOMATIC,
                                        GTK_POLICY_AUTOMATIC );
	gtk_scrolled_window_set_shadow_type( GTK_SCROLLED_WINDOW( sw ),
						GTK_SHADOW_IN );
	priv->view = gtk_tree_view_new();
	screem_tag_tree_set_model( tag_tree, screem_file_browser_get_model( priv->browser ) );
	
        gtk_tree_view_set_rules_hint( GTK_TREE_VIEW( priv->view ), TRUE );

	selection = gtk_tree_view_get_selection( GTK_TREE_VIEW( priv->view ) );
	g_signal_connect( G_OBJECT( selection ), "changed",
			  G_CALLBACK( screem_tagtree_sel_change ), tag_tree );

	priv->paths = g_object_get_data( G_OBJECT( priv->browser ), "paths" );
	
	prend = gtk_cell_renderer_pixbuf_new();
	rend = gtk_cell_renderer_text_new();
	col = gtk_tree_view_column_new();
	
	gtk_tree_view_column_pack_start( col, prend, FALSE );
	gtk_tree_view_column_pack_start( col, rend, TRUE );
	gtk_tree_view_column_set_resizable( col, TRUE );
	gtk_tree_view_append_column( GTK_TREE_VIEW( priv->view ), col );
	
	gtk_tree_view_column_set_title( col, _( "Resources" ) );
	gtk_tree_view_column_set_attributes( col, rend,
					     "text", 
					     FILE_BROWSER_NAME_COL,
					     NULL );
	gtk_tree_view_column_set_attributes( col, prend,
					     "pixbuf",
					     FILE_BROWSER_ICON_COL,
					     NULL );

	gtk_tree_view_set_headers_visible( GTK_TREE_VIEW( priv->view ), TRUE );

	gtk_widget_show( priv->view );

	g_signal_connect( G_OBJECT( priv->view ), "row_activated",
			  G_CALLBACK( screem_tag_tree_click ),
			  tag_tree );

	g_signal_connect( G_OBJECT( priv->view ), "button_press_event",
			  G_CALLBACK( screem_tag_tree_press ),
			  tag_tree );

	gtk_container_add( GTK_CONTAINER( sw ), priv->view );

	gtk_box_pack_start( GTK_BOX( priv->box ), sw, TRUE, TRUE, 0 );


	group = egg_action_group_new( "ScreemTagTreeActions" );
	/* set user data field in action table */
	for( i = 0; i < action_n_entries; ++ i ) {
		action_entries[ i ].user_data = tag_tree;
	}
	egg_action_group_add_actions( EGG_ACTION_GROUP( group ),
					action_entries,
					action_n_entries );
	merge = egg_menu_merge_new ();
	egg_menu_merge_insert_action_group( merge, 
					    EGG_ACTION_GROUP( group ), 
					    0 );
	g_signal_connect( merge, "add_widget", G_CALLBACK( add_widget ), tag_tree );
  	egg_menu_merge_add_ui_from_file( merge, 
					 UIDATADIR"/screem-tag-tree-bar.xml", 
					 NULL );

	gtk_widget_show_all( GTK_WIDGET( tag_tree ) );
}

static void screem_tag_tree_finalize( GObject *object )
{
	ScreemTagTree *tree;
	ScreemTagTreePrivate *priv;
	
	tree = SCREEM_TAG_TREE( object );
	priv = tree->private;
	
	G_OBJECT_CLASS( parent_class )->finalize( object );
}

static void screem_tag_tree_set_prop( GObject *object, guint property_id,
				    const GValue *value, GParamSpec *pspec )
{
	ScreemTagTree *tree;

	tree = SCREEM_TAG_TREE( object );

	switch( property_id ) {
	case ARG_WINDOW:
		tree->private->window = g_value_get_pointer( value );

		break;
	}
}

static void screem_tag_tree_get_prop( GObject *object, guint property_id,
				    GValue *value, GParamSpec *pspec)
{
	switch( property_id ) {
	case ARG_WINDOW:
		g_value_set_pointer( value,
				     (gpointer)
				     SCREEM_TAG_TREE( object )->private->window );
		break;
	}
}

static void screem_tag_tree_size_request( GtkWidget *widget,
                           	  	  GtkRequisition *requisition )
{
	GtkBin *bin;

	bin = GTK_BIN (widget);

	requisition->width = GTK_CONTAINER( widget )->border_width * 2;
	requisition->height = GTK_CONTAINER( widget )->border_width * 2;

	if( bin->child && GTK_WIDGET_VISIBLE( bin->child ) ) {
		GtkRequisition child_requisition;

		gtk_widget_size_request( bin->child, &child_requisition );
		requisition->width += child_requisition.width;
		requisition->height += child_requisition.height;
	}
}

static void screem_tag_tree_size_allocate( GtkWidget *widget,
                              		   GtkAllocation *allocation )
{
	GtkBin *bin;
	GtkAllocation child_allocation;

	bin = GTK_BIN( widget );
	widget->allocation = *allocation;
	
	if( bin->child ) {
	child_allocation.x = allocation->x + 
				GTK_CONTAINER( widget )->border_width; 
	child_allocation.y = allocation->y + 
				GTK_CONTAINER (widget)->border_width;
	child_allocation.width = MAX( allocation->width - 
				      GTK_CONTAINER( widget )->border_width * 2,
					0);
	child_allocation.height = MAX( allocation->height - 
				       GTK_CONTAINER (widget)->border_width * 2,
					0);
	gtk_widget_size_allocate( bin->child, &child_allocation );
	}
}

GType screem_tag_tree_get_type()
{
	static GType type = 0;
	
	if( ! type ) {
		static const GTypeInfo info = {
			sizeof( ScreemTagTreeClass ),
			NULL, /* base init */
			NULL, /* base finalise */
			(GClassInitFunc)screem_tag_tree_class_init,
			NULL, /* class finalise */
			NULL, /* class data */
			sizeof( ScreemTagTree ),
			0, /* n_preallocs */
			(GInstanceInitFunc)screem_tag_tree_init
		};

		type = g_type_register_static( PARENT_TYPE,
					       "ScreemTagTree",
					       &info, 0 );
	}

	return type;
}
